//
//  Renderer.m
//  RenderingWithMetal
//
//  Created by Harold Serrano on 5/7/22.
//

#import <simd/simd.h>
#import "Renderer.h"

// Include header shared between C code here, which executes Metal API commands, and .metal files
#import "ShaderTypes.h"

static float quadVertexData[] =
{
    0.5, -0.5, 0.0, 1.0,
    -0.5, -0.5, 0.0, 1.0,
    -0.5,  0.5, 0.0, 1.0,

    0.5,  0.5, 0.0, 1.0,
    0.5, -0.5, 0.0, 1.0,
    -0.5,  0.5, 0.0, 1.0
};

@implementation Renderer
{
    
    id<MTLDevice> mtlDevice;
    id<MTLCommandQueue> mtlCommandQueue;
    id<MTLRenderPipelineState> renderPipelineState;
    id<MTLBuffer> vertexBuffer;
    id<MTLBuffer> uniformBuffer;
    matrix_float4x4 projectionMatrix;
    float rotation;
    
}

-(nonnull instancetype)initWithMetalKitView:(nonnull MTKView *)view;
{
    self = [super init];
    if(self)
    {
        //Step 1. create a gpu device
        mtlDevice=view.device;
        
        //Step 2. create a command queue
        mtlCommandQueue=[mtlDevice newCommandQueue];
        
        //Step 3. create the library
        id<MTLLibrary> mtlLibrary=[mtlDevice newDefaultLibrary];
        
        //Step 4. create a vertex and fragment function shader
        id<MTLFunction> vertexProgram=[mtlLibrary newFunctionWithName:@"vertexShader"];
        id<MTLFunction> fragmentProgram=[mtlLibrary newFunctionWithName:@"fragmentShader"];
        
        
        //tell gpu our data organization
        MTLVertexDescriptor *vertexDesc=[[MTLVertexDescriptor alloc] init];
        vertexDesc.attributes[0].format=MTLVertexFormatFloat4;
        vertexDesc.attributes[0].bufferIndex=0;
        vertexDesc.attributes[0].offset=0;
        
        
        //stride
        vertexDesc.layouts[0].stride=4*sizeof(float);
        vertexDesc.layouts[0].stepFunction=MTLVertexStepFunctionPerVertex;
        
        //Step 5. Build the rendering pipeline
        MTLRenderPipelineDescriptor *mtlRenderPipelineDescriptor=[[MTLRenderPipelineDescriptor alloc] init];
        
        //assign the vertex and fragment to the descriptor
        [mtlRenderPipelineDescriptor setVertexFunction:vertexProgram];
        [mtlRenderPipelineDescriptor setFragmentFunction:fragmentProgram];
        [mtlRenderPipelineDescriptor setVertexDescriptor:vertexDesc];
        
        //specify the target-texture pixel format
        mtlRenderPipelineDescriptor.colorAttachments[0].pixelFormat=MTLPixelFormatBGRA8Unorm;
        
        //build the rendering pipeline object
        
        renderPipelineState=[mtlDevice newRenderPipelineStateWithDescriptor:mtlRenderPipelineDescriptor error:nil];
        
        //Step 6. Create buffer object and load data into it
        vertexBuffer=[mtlDevice newBufferWithBytes:quadVertexData length:sizeof(quadVertexData) options:MTLResourceOptionCPUCacheModeDefault];
        
        uniformBuffer=[mtlDevice newBufferWithLength:sizeof(Uniforms) options:MTLResourceStorageModeShared];
        
    }

    return self;
}


- (void)drawInMTKView:(nonnull MTKView *)view
{
    //update the transformation
    Uniforms uniforms;

    uniforms.projectionMatrix = projectionMatrix;

    vector_float3 rotationAxis = {0, 1, 0};
    matrix_float4x4 modelMatrix = matrix4x4_rotation(rotation, rotationAxis);
    matrix_float4x4 viewMatrix = matrix4x4_translation(0.0, 0.0, -4.0);

    uniforms.modelViewMatrix = matrix_multiply(viewMatrix, modelMatrix);
    
    memcpy(uniformBuffer.contents, (void*)&uniforms, sizeof(Uniforms));
    
    rotation += .01;
    
    
    
    //Step 8. Create a render pass descriptor
    MTLRenderPassDescriptor *mtlRenderPassDescriptor=view.currentRenderPassDescriptor;
    
    //set the following states for the pipeline
    mtlRenderPassDescriptor.colorAttachments[0].loadAction=MTLLoadActionClear;
    mtlRenderPassDescriptor.colorAttachments[0].clearColor=MTLClearColorMake(1.0, 1.0, 1.0, 1.0);
    mtlRenderPassDescriptor.colorAttachments[0].storeAction=MTLStoreActionStore;
    
    //Step 9. create a command buffer
    id<MTLCommandBuffer> mtlCommandBuffer=[mtlCommandQueue commandBuffer];
    
    //Step 10. Create a command encoder
    id<MTLRenderCommandEncoder> renderEncoder=[mtlCommandBuffer renderCommandEncoderWithDescriptor:mtlRenderPassDescriptor];
    
    //Configure encoder with the pipeline
    [renderEncoder setRenderPipelineState:renderPipelineState];
    
    //set the vertex buffer object
    [renderEncoder setVertexBuffer:vertexBuffer offset:0 atIndex:0];
    [renderEncoder setVertexBuffer:uniformBuffer offset:0 atIndex:1];
    
    //set the draw command
    [renderEncoder drawPrimitives:MTLPrimitiveTypeTriangle vertexStart:0 vertexCount:6];
    
    //end encoding
    
    [renderEncoder endEncoding];
    
    //Step 11. Present the drawable
    [mtlCommandBuffer presentDrawable:view.currentDrawable];
    
    //Step 12. commit the buffer
    [mtlCommandBuffer commit];
    
}

- (void)mtkView:(nonnull MTKView *)view drawableSizeWillChange:(CGSize)size
{
    /// Respond to drawable size or orientation changes here

    float aspect = size.width / (float)size.height;
    projectionMatrix = matrix_perspective_right_hand(65.0f * (M_PI / 180.0f), aspect, 0.1f, 100.0f);
}

#pragma mark Matrix Math Utilities

matrix_float4x4 matrix4x4_translation(float tx, float ty, float tz)
{
    return (matrix_float4x4) {{
        { 1,   0,  0,  0 },
        { 0,   1,  0,  0 },
        { 0,   0,  1,  0 },
        { tx, ty, tz,  1 }
    }};
}

static matrix_float4x4 matrix4x4_rotation(float radians, vector_float3 axis)
{
    axis = vector_normalize(axis);
    float ct = cosf(radians);
    float st = sinf(radians);
    float ci = 1 - ct;
    float x = axis.x, y = axis.y, z = axis.z;

    return (matrix_float4x4) {{
        { ct + x * x * ci,     y * x * ci + z * st, z * x * ci - y * st, 0},
        { x * y * ci - z * st,     ct + y * y * ci, z * y * ci + x * st, 0},
        { x * z * ci + y * st, y * z * ci - x * st,     ct + z * z * ci, 0},
        {                   0,                   0,                   0, 1}
    }};
}

matrix_float4x4 matrix_perspective_right_hand(float fovyRadians, float aspect, float nearZ, float farZ)
{
    float ys = 1 / tanf(fovyRadians * 0.5);
    float xs = ys / aspect;
    float zs = farZ / (nearZ - farZ);

    return (matrix_float4x4) {{
        { xs,   0,          0,  0 },
        {  0,  ys,          0,  0 },
        {  0,   0,         zs, -1 },
        {  0,   0, nearZ * zs,  0 }
    }};
}

@end
